@@ -33,7 +33,7 @@ import statereason
 from statereason import StateReason
 statereason.set_gettext_function (_)
 
-CONNECTING_TIMEOUT = 10 # seconds
+CONNECTING_TIMEOUT = 60 # seconds
 MIN_REFRESH_INTERVAL = 1 # seconds
 
 def state_reason_is_harmless (reason):
@@ -47,7 +47,11 @@ def state_reason_is_harmless (reason):
 
 def collect_printer_state_reasons (connection):
     result = {}
-    printers = connection.getPrinters ()
+    try:
+        printers = connection.getPrinters ()
+    except cups.IPPError:
+        return result
+
     for name, printer in printers.iteritems ():
         reasons = printer["printer-state-reasons"]
         if type (reasons) != list:
@@ -117,7 +121,7 @@ class Monitor:
     DBUS_IFACE="com.redhat.PrinterSpooler"
 
     def __init__(self, watcher, bus=None, my_jobs=True, specific_dests=None,
-                 monitor_jobs=True):
+                 monitor_jobs=True, host=None, port=None, encryption=None):
         self.watcher = watcher
         self.my_jobs = my_jobs
         self.specific_dests = specific_dests
@@ -126,11 +130,22 @@ class Monitor:
         self.printer_state_reasons = {}
         self.printers = set()
 
+        if host:
+            cups.setServer (host)
+        if port:
+            cups.setPort (port)
+        if encryption:
+            cups.setEncryption (encryption)
+        self.host = cups.getServer ()
+        self.port = cups.getPort ()
+        self.encryption = cups.getEncryption ()
+
         self.which_jobs = "not-completed"
         self.reasons_seen = {}
         self.connecting_timers = {}
         self.still_connecting = set()
         self.connecting_to_device = {}
+        self.received_any_dbus_signals = False
 
         if bus == None:
             bus = dbus.SystemBus ()
@@ -149,7 +164,17 @@ class Monitor:
     def cleanup (self):
         if self.sub_id != -1:
             try:
-                c = cups.Connection ()
+                try:
+                    c = cups.Connection (host=self.host,
+                                         port=self.port,
+                                         encryption=self.encryption)
+                except TypeError:
+                    # Parameters to Connection() requires pycups >= 1.9.40.
+                    cups.setServer (self.host)
+                    cups.setPort (self.port)
+                    cups.setEncryption (self.encryption)
+                    c = cups.Connection ()
+
                 c.cancelSubscription (self.sub_id)
                 debugprint ("Canceled subscription %d" % self.sub_id)
             except:
@@ -270,7 +295,8 @@ class Monitor:
                     else:
                         # Don't notify about this, as it must be stale.
                         debugprint ("Ignoring stale connecting-to-device")
-                        debugprint (pprint.pformat (printer_jobs))
+                        if get_debugging ():
+                            debugprint (pprint.pformat (printer_jobs))
 
         self.update_connecting_devices (printer_jobs)
         items = self.reasons_seen.keys ()
@@ -284,7 +310,16 @@ class Monitor:
     def get_notifications(self):
         debugprint ("get_notifications")
         try:
-            c = cups.Connection ()
+            try:
+                c = cups.Connection (host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
+            except TypeError:
+                # Parameters to Connection() requires pycups >= 1.9.40.
+                cups.setServer (self.host)
+                cups.setPort (self.port)
+                cups.setEncryption (self.encryption)
+                c = cups.Connection ()
 
             try:
                 try:
@@ -318,7 +353,8 @@ class Monitor:
             self.sub_seq = seq
             nse = event['notify-subscribed-event']
             debugprint ("%d %s %s" % (seq, nse, event['notify-text']))
-            debugprint (pprint.pformat (event))
+            if get_debugging ():
+                debugprint (pprint.pformat (event))
             if nse.startswith ('printer-'):
                 # Printer events
                 name = event['printer-name']
@@ -387,14 +423,17 @@ class Monitor:
                 deferred_calls.append ((self.watcher.job_added,
                                         (self, jobid, nse, event,
                                          jobs[jobid].copy ())))
-            elif nse == 'job-completed':
-                try:
-                    del jobs[jobid]
-                    deferred_calls.append ((self.watcher.job_removed,
-                                            (self, jobid, nse, event)))
-                except KeyError:
-                    pass
-                continue
+            elif (nse == 'job-completed' or
+                  (nse == 'job-state-changed' and
+                   event['job-state'] == cups.IPP_JOB_COMPLETED)):
+                if not (self.which_jobs in ['completed', 'all']):
+                    try:
+                        del jobs[jobid]
+                        deferred_calls.append ((self.watcher.job_removed,
+                                                (self, jobid, nse, event)))
+                    except KeyError:
+                        pass
+                    continue
 
             try:
                 job = jobs[jobid]
@@ -416,13 +455,33 @@ class Monitor:
         for (fn, args) in deferred_calls:
             fn (*args)
 
+        # Update again when we're told to.  If we're getting CUPS
+        # D-Bus signals, however, rely on those instead.
+        if not self.received_any_dbus_signals:
+            gobject.source_remove (self.update_timer)
+            interval = 1000 * notifications['notify-get-interval']
+            self.update_timer = gobject.timeout_add (interval,
+                                                     self.get_notifications)
+
         return False
 
-    def refresh(self):
+    def refresh(self, which_jobs=None, refresh_all=True):
         debugprint ("refresh")
 
+        if which_jobs != None:
+            self.which_jobs = which_jobs
+
         try:
-            c = cups.Connection ()
+            try:
+                c = cups.Connection (host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
+            except TypeError:
+                # Parameters to Connection() requires pycups >= 1.9.40.
+                cups.setServer (self.host)
+                cups.setPort (self.port)
+                cups.setEncryption (self.encryption)
+                c = cups.Connection ()
         except RuntimeError:
             self.watcher.cups_connection_error (self)
             return
@@ -448,7 +507,6 @@ class Monitor:
             events.extend (["job-created",
                             "job-completed",
                             "job-stopped",
-                            "job-progress",
                             "job-state-changed"])
 
         try:
@@ -460,12 +518,22 @@ class Monitor:
                                                  self.get_notifications)
         debugprint ("Created subscription %d" % self.sub_id)
 
+        if self.monitor_jobs:
+            jobs = self.jobs.copy ()
+            if self.which_jobs not in ['all', 'completed']:
+                # Filter out completed jobs.
+                filtered = {}
+                for jobid, job in jobs.iteritems ():
+                    if job['job-state'] < cups.IPP_JOB_CANCELED:
+                        filtered[jobid] = job
+                jobs = filtered
+
+            self.fetch_first_job_id = 1
+            gobject.timeout_add (5, self.fetch_jobs, refresh_all)
+        else:
+            jobs = {}
+
         try:
-            if self.monitor_jobs:
-                jobs = c.getJobs (which_jobs=self.which_jobs,
-                                  my_jobs=self.my_jobs)
-            else:
-                jobs = {}
             self.printer_state_reasons = collect_printer_state_reasons (c)
             dests = c.getDests ()
             printers = set()
@@ -498,6 +566,85 @@ class Monitor:
         self.jobs = jobs
         return False
 
+    def fetch_jobs (self, refresh_all):
+        try:
+            try:
+                c = cups.Connection (host=self.host,
+                                     port=self.port,
+                                     encryption=self.encryption)
+            except TypeError:
+                # Parameters to Connection() requires pycups >= 1.9.40.
+                cups.setServer (self.host)
+                cups.setPort (self.port)
+                cups.setEncryption (self.encryption)
+                c = cups.Connection ()
+        except RuntimeError:
+            self.watcher.cups_connection_error (self)
+            return False
+
+        limit = 1
+        try:
+            fetched = c.getJobs (which_jobs=self.which_jobs,
+                                 my_jobs=self.my_jobs,
+                                 first_job_id=self.fetch_first_job_id,
+                                 limit=limit)
+        except TypeError:
+            # limit requires pycups >= 1.9.42
+            fetched = c.getJobs (which_jobs=self.which_jobs,
+                                 my_jobs=self.my_jobs)
+            limit = len (fetched) + 1
+        except cups.IPPError, (e, m):
+            self.watcher.cups_ipp_error (self, e, m)
+            return False
+
+        got = len (fetched)
+        debugprint ("Got %s jobs, asked for %s" % (got, limit))
+
+        deferred_calls = []
+        jobs = self.jobs.copy ()
+        jobids = fetched.keys ()
+        jobids.sort ()
+        if got > 0:
+            last_jobid = jobids[got - 1]
+        else:
+            last_jobid = self.fetch_first_job_id + limit
+        for jobid in xrange (self.fetch_first_job_id, last_jobid + 1):
+            try:
+                job = fetched[jobid]
+                if jobs.has_key (jobid):
+                    fn = self.watcher.job_event
+                else:
+                    fn = self.watcher.job_added
+
+                jobs[jobid] = job
+                deferred_calls.append ((fn,
+                                        (self, jobid, '', {}, job.copy ())))
+            except KeyError:
+                # No job by that ID.
+                if jobs.has_key (jobid):
+                    del jobs[jobid]
+                    deferred_calls.append ((self.watcher.job_removed,
+                                            (self, jobid, '', {})))
+
+        self.update (jobs)
+        self.jobs = jobs
+
+        for (fn, args) in deferred_calls:
+            fn (*args)
+
+        if got < limit:
+            # That's all.  Don't run this timer again.
+            return False
+
+        # Remember where we got up to and run this timer again.
+        next = jobid + 1
+
+        while not refresh_all and self.jobs.has_key (next):
+            next += 1
+
+        self.fetch_first_job_id = next
+        return True
+
     def sort_jobs_by_printer (self, jobs=None):
         if jobs == None:
             jobs = self.jobs
@@ -528,6 +675,8 @@ class Monitor:
     def handle_dbus_signal(self, *args):
         gobject.source_remove (self.update_timer)
         self.update_timer = gobject.timeout_add (200, self.get_notifications)
+        if not self.received_any_dbus_signals:
+            self.received_any_dbus_signals = True
 
 if __name__ == '__main__':
     set_debugging (True)
